
from muteria.drivers.criteria.tools_by_languages.c.gpmutation.gpmutation \
                                                import CriteriaToolGPMutation 


StaticCriteriaTool = CriteriaToolGPMutation
